$wnd.rstudio.runAsyncCallback15("function dEq(){}\nfunction bEq(a){this.a=a}\nfunction gEq(a){this.a=a}\nfunction iEq(a){this.a=a}\nfunction BDq(a){this.a=a}\nfunction mFq(a){this.a=a}\nfunction ZVi(a){this.a=a}\nfunction zDq(a){_zi.call(this,a)}\nfunction iDq(a){sDq(a,new gEq(a))}\nfunction d2e(c,a,b){return c.slice(a,b)}\nfunction c2e(a,b){return d2e(a,b,a.length||0)}\nfunction GUi(a,b){STi();(kQe(b),b)&&rDq(a.a,jRB,'')}\nfunction pDq(a,b){_2h(a.d,new adr(b,true,true),false)}\nfunction mDq(a){N$t(a.o,new bEq(a),new dEq,cDq.u8b())}\nfunction nDq(a){vUi(a.b,cDq.B8b(),cDq.B8b(),new iEq(a))}\nfunction oDq(a){a.K.M7b();gn((_m(),$m),new BDq(a))}\nfunction sDq(a,b){a.c!=null?fEq(b,a.c):Tym(a.k,new DDq(a,b))}\nfunction DDq(a,b){Kuf();this.a=a;this.b=b;Muf.call(this)}\nfunction fEq(a,b){pDq(a.a,'devtools::load_all(\"'+b+'\")')}\nfunction Tym(a,b){pDm(a,Tvz,'devtools_load_all_path',new xL,b)}\nfunction vUi(a,b,c,d){gUi(a,b,c,yWi(qZo(a.j.b),'roxygen'),VTi(a,'roxygen'),false,new ZVi(d))}\nfunction fDq(a,b){var c,d;HMi(a.s.g,null);d=b.outputs;(d.length||0)>zty&&(d=c2e(d,(d.length||0)-zty));for(c=0;c<(d.length||0);c++)DCq(a.s,d[c]||null);b.errors.length>0&&CCq(a.s,b.errors_base_dir,b.errors,false,0,true,b.type);b.running?U9e(a.a.Wj,true,false):GMi(a.s.g);PTg(a.s.g.c.g)}\nfunction iFq(a){try{if(a.c<0)return;if(!a.a.a){a.a.a=mHq(a.a.b.a);Qnq(a.a,a.a.a)}switch(a.c){case 1:fDq(a.a.a,a.b[0]);break;case 2:a.a.a.K.K7b();break;case 3:a.a.a.K.L7b();break;case 4:rDq(a.a.a,v_A,'');break;case 5:rDq(a.a.a,'build-binary-package','');break;case 6:rDq(a.a.a,'build-full','');break;case 7:rDq(a.a.a,'build-incremental','');break;case 8:rDq(a.a.a,'build-source-package','');break;case 9:rDq(a.a.a,'check-package','');break;case 10:rDq(a.a.a,'clean-all','');break;case 11:mDq(a.a.a);break;case 12:nDq(a.a.a);break;case 13:oDq(a.a.a);break;case 14:qEm(a.a.a.k,new zDq((dDq(),cDq).K8b()));break;case 15:rDq(a.a.a,'test-package','');break;case 18:a.a.a.K.N7b();}}finally{a.c<=0&&a.b[0]!=null&&a.b[0].mn()}}\nvar jRB='roxygenize-package',kRB='Error Terminating Build',lRB='Unable to terminate build. Please try again.',mRB='Building package documentation',nRB='Terminating Build...';SNd(2336,1,GGy,ZVi);_.qw=function $Vi(a){GUi(this.a,a)};var Vyb=JAe(2336);SNd(8202,394,{},zDq);_._m=function ADq(a){(kQe(a),a)||mzi(3,(dDq(),cDq).x8b(),cDq.y8b()).cV()};var wic=JAe(8202);SNd(8203,1,LGy,BDq);_.mn=function CDq(){PTg(this.a.s.g.c.g)};var xic=JAe(8203);SNd(8204,54,{},DDq);_.BA=function EDq(a){this.a.c=a;fEq(this.b,this.a.c)};var yic=JAe(8204);SNd(8195,1,LGy,bEq);_.mn=function cEq(){iDq(this.a)};var Mic=JAe(8195);SNd(8196,1,LGy,dEq);_.mn=function eEq(){dDq()};var Nic=JAe(8196);SNd(8194,1,GGy,gEq);_.qw=function hEq(a){fEq(this,a)};var Oic=JAe(8194);SNd(8197,1,LGy,iEq);_.mn=function jEq(){rDq(this.a,jRB,'')};var Pic=JAe(8197);SNd(255,1,Tez);_.kn=function lFq(){iFq((new mFq(this)).a)};SNd(7470,1,LGy,mFq);_.mn=function nFq(){iFq(this.a)};var kjc=JAe(7470);SNd(5605,1,x_A);_.x8b=function zFq(){return kRB};_.y8b=function AFq(){return lRB};_.B8b=function DFq(){return mRB};_.K8b=function MFq(){return nRB};SNd(5604,1,x_A);_.x8b=function YFq(){return kRB};_.y8b=function ZFq(){return lRB};_.B8b=function aGq(){return mRB};_.K8b=function jGq(){return nRB};SNd(5603,1,x_A);_.x8b=function vGq(){return 'Erreur de fin de construction'};_.y8b=function wGq(){return 'Impossible de terminer la construction. Veuillez r\\xE9essayer.'};_.B8b=function zGq(){return 'Documentation du paquet de construction'};_.K8b=function IGq(){return 'Terminer la construction...'};Dqy(hm)(15);\n//# sourceURL=rstudio-15.js\n")
