$wnd.rstudio.runAsyncCallback17("function gUt(a){this.a=a}\nfunction E3t(a){this.a=a}\nfunction G3t(a,b){this.a=a;this.b=b}\nfunction JKv(a,b){this.a=a;this.b=b}\nfunction dLv(a,b){this.a=a;this.b=b}\nfunction N$w(a,b){M$w();this.b=a;this.a=b}\nfunction U$t(a,b){AZt();var c;c=a;Ywv(c,b.a)}\nfunction rJw(a,b){Osk(a.w,b)}\nfunction Qwv(a,b){rJw(a.vb,b)}\nfunction uSt(a){tzi(a.o,a.c,null)}\nfunction vSt(a){tzi(a.o,a.g,null)}\nfunction aTt(a){rTt(a.p,a.g,SNA,null)}\nfunction fTt(a){rTt(a.p,a.g,'next',null)}\nfunction hTt(a){rTt(a.p,a.g,'prev',null)}\nfunction iTt(a){rTt(a.p,a.g,'refresh',null)}\nfunction wSt(a){tzi(a.o,a.g+'?print-pdf',null)}\nfunction Duj(a){return {items:a}}\nfunction Jwv(a,b){gn((_m(),$m),new dLv(a,b))}\nfunction T$t(a,b){AZt();gn((_m(),$m),new G3t(b,a))}\nfunction tUh(a){$wnd.desktop.setPresentationUrl(a)}\nfunction Osk(a,b){a.d.navigateToPresentationEditorLocation(b)}\nfunction Ywv(a,b){F2w(a.A,_sz,false)?fvv(a,new JKv(a,b)):IPv(a.ab,b)}\nfunction oSt(a,b){tCe(b.substr(0,4),lvy)||(b=lzm(a.q,b));return b}\nfunction tSt(a){!!a.b&&_2h(a.n,new N$w(wnf(a.b.source_file),xuj(a.d,a.a,a.e)),false)}\nfunction yuj(a,b){var c,d,e,f;b==-1&&(b=a.auto_slide_level);f=wuj(a,b,true);for(c=0;c<f.length;c++){e=f[c];for(d=0;d<e.items.length;d++){if(tCe(e.items[d].type,O5y))return c}}return 0}\nfunction xuj(a,b,c){var d,e,f,g,h;c==-1&&(c=b.auto_slide_level);h=wuj(b,c,false);if(a<h.length){g=h[a];zl(g.items,Auj(g.items[0].row))}e=[];for(d=0;d<h.length;d++){g=h[d];for(f=0;f<g.items.length;f++){xl(e,g.items[f])}}return zuj(e,b.auto_slide_level)}\nfunction IPv(a,b){var c,d,e,f,g,h,i;e=HPv(a).items;d=[];for(g=0;g<e.length;g++){h=e[g];tCe(h.type,O5y)||(d[d.length]=h,undefined)}c=-1;i=b.items;for(f=0;f<i.length;f++){if(tCe(i[f].type,O5y)){c=f;break}}if(c>=0){h=d[c];!!h&&LUu(a.a,L4w(h.row,0),true)}}\nfunction wuj(a,b,c){var d,e,f,g,h,i;i=[];g=a.items;for(d=0;d<g.length;d++){f=g[d];e=i.length==0||tCe(f.type,Dsy)||tCe(f.type,'hr')||tCe(f.type,asy)&&f.level<=b;if(e){h=Duj([]);xl(h.items,f);i[i.length]=h}else (c||!tCe(f.type,O5y))&&xl(i[i.length-1].items,f)}return i}\nfunction xSt(a,b){var c,d;c=b.a;d=oSt(a,c.url);a.p=yuj(b.a.editor_state,b.a.slide_level);a.p==0&&a.d!=0&&(a.p=a.d);Nuq(a.k);if(!tCe(d,a.g)||!a.k.p){a.g=d;a.c=d;a.b=c.quarto_navigation;a.a=c.editor_state;a.e=c.slide_level;(oTh()||!!$wnd.remoteDesktop)&&tUh(a.g);eTt(a.k,a.g,a.b)}else{iTt(a.k)}}\nfunction eTt(a,b,c){var d;(oTh()||!!$wnd.remoteDesktop)&&(t6e(b)?(Oh(a.d).removeAttribute(n5y),undefined):(Oh(a.d).setAttribute(n5y,eOB),undefined));a.b=b;a.p=null;a.g=null;d=Qh(a.d);d>0&&yi(a,new Tkf(hO(d*0.6666666666666666)+((a.L?Ph(a.L):0)+(a.P?Ph(a.P):0))));VRg(a.d,a.b);c.is_website?Srm(a.j):Rrm(a.j,c.source_file,c.output_file);Bmh(a.q)}\nfunction cUt(a){try{if(a.c<0)return;if(!a.a.a){a.a.a=HUt(a.a.b.a);Qnq(a.a,a.a.a)}switch(a.c){case 1:a.a.a.K.K7b();break;case 2:a.a.a.K.L7b();break;case 3:tSt(a.a.a);break;case 4:aTt(a.a.a.k);break;case 5:fTt(a.a.a.k);break;case 6:uSt(a.a.a);break;case 7:vSt(a.a.a);break;case 8:hTt(a.a.a.k);break;case 9:wSt(a.a.a);break;case 10:xSt(a.a.a,a.b[0]);break;case 11:iTt(a.a.a.k);break;case 12:a.a.a.K.M7b();break;case 13:a.a.a.K.N7b();}}finally{a.c<=0&&a.b[0]!=null&&a.b[0].mn()}}\nSNd(305,1,Tez);_.kn=function fUt(){cUt((new gUt(this)).a)};SNd(7489,1,LGy,gUt);_.mn=function hUt(){cUt(this.a)};var BRc=JAe(7489);SNd(3292,1,GGy,E3t);_.qw=function F3t(a){T$t(this.a,a)};var tVc=JAe(3292);SNd(3291,1,fty,G3t);_.mn=function H3t(){U$t(this.a,this.b)};var uVc=JAe(3291);SNd(5065,1,LGy,JKv);_.mn=function KKv(){Jwv(this.a,this.b)};var K9c=JAe(5065);SNd(5064,1,fty,dLv);_.mn=function eLv(){Qwv(this.a,this.b)};var V9c=JAe(5064);SNd(3340,Ruy,{},N$w);_.$n=function O$w(a){var b,c;b=(c=uaj(a.n,this.b),!!c&&aO(c,76)?c:(oaj(),iaj));E7t(a.b,this.b,b,new E3t(this))};_.ao=function Q$w(){return L$w};_._n=function P$w(){return L$w};var _qd=JAe(3340);Dqy(hm)(17);\n//# sourceURL=rstudio-17.js\n")
